#include "antiglut.h"
#include <math.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>

#include <mvlib.h>

#include <../source/mvlib.c>
#include <../source/rg.c>
#include <../source/agmainloop.c>

#define MAXFIB	100000
#define MAXC	(10 * MAXFIB)

#define	DRAW_SCALE	(vec_t)5e6

#define	CLEAR_COLOR			1.0f,	1.0f,	1.0f,	0.5f
#define	FIB_ALONE_COLOR		0.55f,	0.75f,	0.55f,	0.4f
#define	FIB_CP_COLOR		0.8f,	0.4f,	0.4f,	0.4f
#define	CONTACT_PT_COLOR	0.0f,	0.0f,	0.0f,	0.7f

/*
 * Size of simulation box
 */
#define BOXEXT	(vec_t)1e-6
vec3_t bbox[2] = {{-BOXEXT,-BOXEXT,-BOXEXT},{BOXEXT,BOXEXT,BOXEXT}};

int		num_fibers = 10000;
vec_t	fiber_len = 2e-7;
vec_t	fiber_diameter = 2e-9;

/*
int		nums_fibers[] = {2500,5000,7500,10000,15000,20000,25000,30000,40000};
*/
int		nums_fibers[] = {7500};
vec_t	fiber_lens[] = {2e-7};
vec_t	fiber_diameters[] = {2e-9};


vec3_t	fib[MAXFIB][2];

typedef struct fc_info_s
{
	int fn1, fn2;	/* indices of contacting fibers */
	vec_t dist;		/* minimal distance between fibers */
	vec3_t cp;		/* contact point */
}	fc_info_t, *fc_info_p;

int	num_contacts = 0;
fc_info_t fc[MAXC];

static ran_state_t state;

static void
fill_box_with_fibers()
{
	int i;
	vec_t hlen = 0.5f * fiber_len;
	vec3_t fo, halfdir;
	for(i = 0;i < num_fibers;i++)
	{
		ranNextVec3Range(&state,fo,-BOXEXT,BOXEXT);
		ranDirection(&state,halfdir);
		vec3Scale(halfdir,halfdir,hlen);
		vec3Sub(fib[i][0],fo,halfdir);
		vec3Add(fib[i][1],fo,halfdir);
	}
}

/* volumetric rate of fibers */
static vec_t fiber_vrate()
{
	vec_t fvol, vol;
	vec3_t bs;
	vec3Sub(bs, bbox[1], bbox[0]);
	vol = bs[0] * bs[1] * bs[2];
	fvol = num_fibers * M_PI * fiber_diameter * fiber_diameter * (vec_t)0.25 * fiber_len;
	return fvol / vol;
}

static int
check_for_contacts()
{
	int i, j;
	vec3_t cp;
	vec_t cd;
	num_contacts = 0;
	for(i = 0;i < num_fibers;i++)
		for(j = i + 1;j < num_fibers;j++)
		{
			if(!segCrossing(fib[i][0],fib[i][1],fib[j][0],fib[j][1],&cd,cp))
				continue;
			if(cd > fiber_diameter)
				continue;
			if(num_contacts >= MAXC)
			{
				puts("Too many contacts!");
				return num_contacts;
			}
			fc[num_contacts].fn1 = i;
			fc[num_contacts].fn2 = j;
			fc[num_contacts].dist = cd;
			vec3Copy(fc[num_contacts].cp, cp);
			num_contacts++;
		}
	return num_contacts;
}

void onerun()
{
	char *ptr, buf[1024];
	fill_box_with_fibers();
	check_for_contacts();
	sprintf(buf,"%i;%g;%g;%g;%i",
			num_fibers,
			fiber_len,fiber_diameter,
			fiber_vrate(),
			num_contacts);
	for(ptr = buf;*ptr;ptr++)
		if(*ptr == '.')
			*ptr = ',';
	puts(buf);
	fflush(stdout);
}

void longrun_generic(void (*onerunproc)(), vboolean print_header)
{
	int i, j, k;
	int im = sizeof(nums_fibers) / sizeof(int);
	int jm = sizeof(fiber_lens) / sizeof(vec_t);
	int km = sizeof(fiber_diameters) / sizeof(vec_t);
	printf("Starting longrun: %i total iterations\n", im * jm * km);
	for(i = 0;i < im;i++)
	{
		num_fibers = nums_fibers[i];
		assert(num_fibers <= MAXFIB);
		for(j = 0;j < jm;j++)
		{
			fiber_len = fiber_lens[j];
			if(print_header)
				puts("num_fibers;fiber_len;fiber_diameter;fiber_vrate;num_contacts");
			for(k = 0;k < km;k++)
			{
				fiber_diameter = fiber_diameters[k];
				onerunproc();
			}
			/* puts("INNER COMPLETE"); */
		}
		/* puts("MID COMPLETE"); */
	}
	/* puts("OUTER COMPLETE"); */
}

void longrun()
{
	longrun_generic(onerun, true);
}

vboolean draw_main(void)
{
	int i;
	glPushMatrix();
	glScalef(DRAW_SCALE, DRAW_SCALE, DRAW_SCALE);

	glColor4f(.3f,0.2f,.25f,.5f);
	glLineWidth(1.8f);
	bboxDraw(bbox[0], bbox[1]);

	glLineWidth(1.2f);
	glColor4f(FIB_ALONE_COLOR);
	glBegin(GL_LINES);
	for(i = 0;i < num_fibers;i++)
	{
		glVertex3fv(fib[i][0]);
		glVertex3fv(fib[i][1]);
	}
	glEnd();

	glPointSize(3.5f);
	glColor4f(CONTACT_PT_COLOR);
	glBegin(GL_POINTS);
	for(i = 0;i < num_contacts;i++)
		glVertex3fv(fc[i].cp);
	glEnd();
	glLineWidth(1.4f);
	glColor4f(FIB_CP_COLOR);
	glBegin(GL_LINES);
	for(i = 0;i < num_contacts;i++)
	{
		glVertex3fv(fib[fc[i].fn1][0]);
		glVertex3fv(fib[fc[i].fn1][1]);
		glVertex3fv(fib[fc[i].fn2][0]);
		glVertex3fv(fib[fc[i].fn2][1]);
	}
	glEnd();
	
	glPopMatrix();
	return true;
}

void onerun_draw()
{
	double vf, rnc, pt;
	char title[512];
	fill_box_with_fibers();
	check_for_contacts();
	vf = fiber_vrate();
	rnc = (double)num_contacts / (double)num_fibers;
	pt = vf / rnc;
	printf("num_fibers\t%i\n"
			"length\t%g\n"
			"diameter\t%g\n"
			"vf\t%g\n"
			"number of contacts\t%i\n"
			"relative NC\t%g\n"
			"treshold\t%g\n",
			num_fibers,
			fiber_len, fiber_diameter,
			vf,
			num_contacts,
			rnc,
			pt);
	sprintf(title, "Form factor %u, vf = %.2g (Nf %u), relative NC = %.3f, app. treshold = %.3f", (unsigned)(fiber_len / fiber_diameter), vf, num_fibers, rnc, pt);
	agWindowConfig(title,800,800,GL_FALSE);
	if(!agInit(AG_REQUIRED_1_1,GL_FALSE,GL_FALSE,GL_FALSE))
		return;
	glClearColor(CLEAR_COLOR);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_BLEND);
	glEnable(GL_LINE_SMOOTH);
	glEnable(GL_POINT_SMOOTH);
	agMainLoop(draw_main, NULL);
	agShutdown();
}

static agm_globals_t new_agm =
{
        {-1,-1,-1}, {1,1,1},
        0, -78, 0, 0,
        0, 0.6f,
        2.2, 66, 650, 3.4,
        GL_FALSE,
        false, true, true, false
};

void longrun_draw()
{
	longrun_generic(onerun_draw, false);
}

int main()
{
	assert(sizeof(GLfloat) == sizeof(vec_t));
	ranInitState(&state,1);
	memcpy(&agm, &new_agm, sizeof(agm));
	longrun_draw();
	return 0;
}
