#include <antiglut.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>

/*
 * K: FIXME
 */
#ifdef __WINDOWS__ /*[*/
#include <io.h>
#else
#include <sys/types.h>
#include <unistd.h>
#endif /*]*/

int printOglError(char *file, int line)
{
    /* Returns 1 if an OpenGL error occurred, 0 otherwise. */
    GLenum glErr;
    int    retCode = 0;

    glErr = glGetError();
    while (glErr != GL_NO_ERROR)
    {
        printf("glError in file %s @ line %d\n", file, line);
        retCode = 1;
        glErr = glGetError();
    }
    return retCode;
}

#define printOpenGLError() printOglError(__FILE__, __LINE__)

struct complete_shader
{
	GLchar *vert_src;
	GLchar *frag_src;
	GLuint vert;
	GLuint frag;
	GLuint prog;
};

/*
 * Get the location of a uniform variable
 */
static
GLint getUniLoc(GLuint program, const GLchar *name)
{
    GLint loc;
    loc = glGetUniformLocation(program, name);
    if (loc == -1)
        printf("No such uniform named \"%s\"\n", name);
    printOpenGLError();  /* Check for OpenGL errors */
    return loc;
}

/*
 * Print out the information log for a shader object
 */
static
void printShaderInfoLog(GLuint shader)
{
    int infologLength = 0;
    int charsWritten  = 0;
    GLchar *infoLog;
    printOpenGLError();  /* Check for OpenGL errors */
    glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &infologLength);
    printOpenGLError();  /* Check for OpenGL errors */
    if (infologLength > 0)
    {
        infoLog = (GLchar *)malloc(infologLength);
        if (infoLog == NULL)
        {
            printf("ERROR: Could not allocate InfoLog buffer\n");
            exit(1);
        }
        glGetShaderInfoLog(shader, infologLength, &charsWritten, infoLog);
        printf("Shader InfoLog:\n%s\n\n", infoLog);
        free(infoLog);
    }
    printOpenGLError();  /* Check for OpenGL errors */
}

/* 
 * Print out the information log for a program object
 */
static
void printProgramInfoLog(GLuint program)
{
    int infologLength = 0;
    int charsWritten  = 0;
    GLchar *infoLog;
    printOpenGLError();  /* Check for OpenGL errors */
    glGetProgramiv(program, GL_INFO_LOG_LENGTH, &infologLength);
    printOpenGLError();  /* Check for OpenGL errors */
    if (infologLength > 0)
    {
        infoLog = (GLchar *)malloc(infologLength);
        if (infoLog == NULL)
        {
            printf("ERROR: Could not allocate InfoLog buffer\n");
            exit(1);
        }
        glGetProgramInfoLog(program, infologLength, &charsWritten, infoLog);
        printf("Program InfoLog:\n%s\n\n", infoLog);
        free(infoLog);
    }
    printOpenGLError();  /* Check for OpenGL errors */
}

static char *
readGLprog(char *filename)
{
    FILE *fh;
    int shader_size;
    char name[256] = "glprogs/";
    int count;
    char *shaderText;
    strcat(name, filename);
    fh = fopen(name, "r");
    if (!fh)
	{
		perror("fopen");
        return NULL;
	}
    fseek(fh, 0, SEEK_END);
    shader_size = ftell(fh);
    fseek(fh, 0, SEEK_SET);
    if(!shader_size)
	{
		printf("Empty shader source \"%s\"\n", name);
		fclose(fh);
        return NULL;
	}
    shaderText = (GLchar *)malloc(shader_size + 1);
    count = (int)fread(shaderText, 1, shader_size, fh);
    if(count != shader_size)
    {
    	fclose(fh);
    	free(shaderText);
    	return NULL;
    }
    shaderText[count] = '\0';
    fclose(fh);
    return shaderText;
}

static void
freeShader(struct complete_shader *shader)
{
	if(shader->prog)
	{
		glDetachShader(shader->prog, shader->vert);
		glDetachShader(shader->prog, shader->frag);
		glDeleteProgram(shader->prog);
		shader->prog = 0;
	}
	if(shader->vert)
	{
		glDeleteShader(shader->vert);
		shader->vert = 0;
	}
	if(shader->frag)
	{
		glDeleteShader(shader->frag);
		shader->frag = 0;
	}
	if(shader->vert_src)
	{
		free(shader->vert_src);
		shader->vert_src = NULL;
	}
	if(shader->frag_src)
	{
		free(shader->frag_src);
		shader->frag_src = NULL;
	}
}

static int
readShader(char *vertname,char *fragname,
	struct complete_shader *shader)
{
    GLint  vertCompiled, fragCompiled;    /* status values */
    GLint  linked;
	shader->vert_src = shader->frag_src = NULL;
	shader->vert = shader->frag = shader->prog = 0;
	if(!(shader->vert_src = readGLprog(vertname)))
		return 0;
	if(!(shader->frag_src = readGLprog(fragname)))
	{
		freeShader(shader);
		return 0;
	}
    shader->vert = glCreateShader(GL_VERTEX_SHADER);
    shader->frag = glCreateShader(GL_FRAGMENT_SHADER);
	glShaderSource(shader->vert, 1, (const GLchar**)&shader->vert_src, NULL);
    glShaderSource(shader->frag, 1, (const GLchar**)&shader->frag_src, NULL);
    /*
     * Compile vertex shader and print out compiler log.
     */
    glCompileShader(shader->vert);
    printOpenGLError();
    glGetShaderiv(shader->vert, GL_COMPILE_STATUS, &vertCompiled);
    printShaderInfoLog(shader->vert);
    /*
     * Compile fragment shader and print out compiler log.
     */
    glCompileShader(shader->frag);
    printOpenGLError();
    glGetShaderiv(shader->frag, GL_COMPILE_STATUS, &fragCompiled);
    printShaderInfoLog(shader->frag);
    if (vertCompiled && fragCompiled)
    {
		shader->prog = glCreateProgram();
		glAttachShader(shader->prog, shader->vert);
		glAttachShader(shader->prog, shader->frag);
		/*
		 * Link the program and print out the info log
		 */
		glLinkProgram(shader->prog);
		printOpenGLError();
		glGetProgramiv(shader->prog, GL_LINK_STATUS, &linked);
		printProgramInfoLog(shader->prog);
		if(linked)
			return 1;
    }
    freeShader(shader);
    return 0;
}

/* ***********************************************************************
   ********************************************************************* */

struct complete_shader brick;
struct complete_shader art;

static int
installShaders()
{
	readShader("brick.vert", "brick.frag", &brick);
	readShader("art.vert", "art.frag", &art);
	if(!brick.prog || !art.prog)
		return 0;
    /* Set up initial uniform values */
    glUseProgram(brick.prog);
    glUniform3f(getUniLoc(brick.prog, "BrickColor"), 0.8f, 0.3f, 0.1f);
    glUniform3f(getUniLoc(brick.prog, "MortarColor"), 0.2f, 0.2f, 0.3f);
    glUniform2f(getUniLoc(brick.prog, "BrickSize"), 0.38f, 0.10f);
    glUniform2f(getUniLoc(brick.prog, "BrickPct"), 0.90f, 0.85f);
    glUniform3f(getUniLoc(brick.prog, "LightPosition"), 0, 0, 4.0f);
    glUseProgram(art.prog);
    glUniform3f(getUniLoc(art.prog, "PhaseShift"), 0.2f, 0.4f, 0.6f);
    glUniform3f(getUniLoc(art.prog, "RadialFreq"), 2.0f, 2.0f, 2.0f);
    glUniform3f(getUniLoc(art.prog, "Center"), 400.0f, 400.0f, 0.0f);
    glUniform3f(getUniLoc(art.prog, "WindowSize"), 800.0f, 800.0f, 0.0f);
    return 1;
}

static void
freeShaders()
{
	freeShader(&brick);
	freeShader(&art);
}

/* ***********************************************************************
	drawing stuff
   ********************************************************************* */

static void drawCube(void)
{
    #define	size 1.0f
/*
    #define scale 0.2f
    #define	delta 0.1f
*/
    #define scale 1.0f
    #define	delta 0.0f

    float A[3] = { size,  size,  size * scale + delta };
    float B[3] = { size,  size, -size * scale + delta };
    float C[3] = { size, -size, -size * scale };
    float D[3] = { size, -size,  size * scale };
    float E[3] = {-size,  size,  size * scale + delta };
    float F[3] = {-size,  size, -size * scale + delta };
    float G[3] = {-size, -size, -size * scale };
    float H[3] = {-size, -size,  size * scale };

    float I[3] = { 1.0f,  0.0f,  0.0f};
    float K[3] = {-1.0f,  0.0f,  0.0f};
    float L[3] = { 0.0f,  0.0f, -1.0f};
    float M[3] = { 0.0f,  0.0f,  1.0f};
    float N[3] = { 0.0f,  1.0f,  0.0f};
    float O[3] = { 0.0f, -1.0f,  0.0f};


    glBegin(GL_QUADS);

    glNormal3fv(I);

    glTexCoord2f(1,1);
    glVertex3fv(D);
    glTexCoord2f(0,1);
    glVertex3fv(C);
    glTexCoord2f(0,0);
    glVertex3fv(B);
    glTexCoord2f(1,0);
    glVertex3fv(A);

    glNormal3fv(K);

    glTexCoord2f(1,1);
    glVertex3fv(G);
    glTexCoord2f(0,1);
    glVertex3fv(H);
    glTexCoord2f(0,0);
    glVertex3fv(E);
    glTexCoord2f(1,0);
    glVertex3fv(F);

    glNormal3fv(L);

    glTexCoord2f(1,1);
    glVertex3fv(C);
    glTexCoord2f(0,1);
    glVertex3fv(G);
    glTexCoord2f(0,0);
    glVertex3fv(F);
    glTexCoord2f(1,0);
    glVertex3fv(B);

    glNormal3fv(M);

    glTexCoord2f(1,1);
    glVertex3fv(H);
    glTexCoord2f(0,1);
    glVertex3fv(D);
    glTexCoord2f(0,0);
    glVertex3fv(A);
    glTexCoord2f(1,0);
    glVertex3fv(E);

    glNormal3fv(N);

    glTexCoord2f(1,1);
    glVertex3fv(E);
    glTexCoord2f(0,1);
    glVertex3fv(A);
    glTexCoord2f(0,0);
    glVertex3fv(B);
    glTexCoord2f(1,0);
    glVertex3fv(F);

    glNormal3fv(O);
    
    glTexCoord2f(1,1);
    glVertex3fv(G);
    glTexCoord2f(0,1);
    glVertex3fv(C);
    glTexCoord2f(0,0);
    glVertex3fv(D);
    glTexCoord2f(1,0);
    glVertex3fv(H);
    
    glEnd();
#undef size
#undef scale
#undef delta

}

#define	REPX	3
#define	REPY	3
#define	REPZ	3

#define	TRANSX	3
#define	TRANSY	4
#define	TRANSZ	-5

static void drawCubes(GLuint *progs, int num_progs)
{
	int i, j, k;
	int prog_idx = 0;
	glPushMatrix();
	glScalef(0.5f, 0.5f, 0.5f);
	glTranslatef(0, 0, -(REPZ - 1) * TRANSZ * 0.5f);
	for(k = 0;k < REPZ;k++)
	{
		glPushMatrix();
		glTranslatef(0, -(REPY - 1) * TRANSY * 0.5f, 0);
		for(j = 0;j < REPY;j++)
		{
			glPushMatrix();
			glTranslatef(-(REPX - 1) * TRANSX * 0.5f, 0, 0);
			for(i = 0;i < REPX;i++)
			{
				glUseProgram(num_progs ? progs[prog_idx++ % num_progs] : 0);
				drawCube();
				glTranslatef(TRANSX, 0, 0);
			}
			glPopMatrix();
			glTranslatef(0, TRANSY, 0);
		}
		glPopMatrix();
		glTranslatef(0, 0, TRANSZ);
	}	
	glPopMatrix();
	
}


static void NextClearColor(void)
{
    static int color = 0;

    switch( color++ )
    {
        case 0:  glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
             break;
        case 1:  glClearColor(0.2f, 0.2f, 0.3f, 1.0f);
             break;
        default: glClearColor(0.7f, 0.7f, 0.7f, 1.0f);
             color = 0;
             break;
    }
}


/* ***********************************************************************
	anti GLUT glue
   ********************************************************************* */

static GLfloat alpha = 0, beta = 0;
static GLfloat deltax = 0, deltay = 0;
static GLfloat fmin = 4 , fcenter = 15, fmax = 26, scale = 2.0f;
static int pofs = -1;

static
void display(void)
{
	GLuint progs[2];

	progs[0] = brick.prog;
	progs[1] = art.prog;
	
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glLoadIdentity();

	glTranslatef(0,0,-fcenter);
	glScalef(scale,scale,scale); 

	glTranslatef(deltax,deltay,0);
	glRotatef(alpha,0,1,0);
	glRotatef(beta,1,0,0);

	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	drawCubes(progs, 2);

	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	glEnable(GL_POLYGON_OFFSET_LINE);
	glPolygonOffset(pofs, -1);
	drawCubes(NULL, 0);
	glDisable(GL_POLYGON_OFFSET_LINE);

	glFlush();
	agSwap();
}

static void
correct_frustum(void)
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glFrustum(-1,1,-1,1,fmin,fmax);
	glMatrixMode(GL_MODELVIEW);
}

GLfloat dpr = 0.2, dpg = 0.2, dpb = 0.6;

void shiftphase(int idx, GLfloat delta)
{
	switch(idx)
	{
		case 1: dpr += delta; break;
		case 2: dpg += delta; break;
		default: dpb += delta; break;
	}
	glUseProgram(art.prog);
	glUniform3f(getUniLoc(art.prog, "PhaseShift"),dpr,dpg,dpb);
}

GLfloat fr = 2, fg = 2, fb = 2;

void scalefreq(int idx, GLfloat scale)
{
	switch(idx)
	{
		case 1: fr *= scale; break;
		case 2: fg *= scale; break;
		default: fb *= scale; break;
	}
	glUseProgram(art.prog);
	glUniform3f(getUniLoc(art.prog, "RadialFreq"),fr,fg,fb);
}

GLfloat center[3] = {400.0, 400.0, 0};

void shiftcenter(int dx, int dy)
{
	glUseProgram(art.prog);
	center[0] += (GLfloat)dx;
	center[1] -= (GLfloat)dy;
	glUniform3f(getUniLoc(art.prog, "Center"),center[0], center[1], center[2]);
}

static int
handle_event_check_break(ui_event_p in)
{
	switch(in->type)
	{
		case ui_key_event:
			switch(in->key.scan)
			{
				case sc_escape:
					return 1;
		        case sc_b:
					if(in->key.press)
			            NextClearColor();
		            break;
		        case sc_a:
					if(in->key.press)
					{
						pofs += 1;
					}
		            break;
		        case sc_s:
					if(in->key.press)
					{
						pofs -= 1;
					}
		            break;
				case sc_wheelup:
					if(in->state & UI_CAPS_LOCK_MASK)
					{
						if(in->state & UI_CONTROL_MASK)
						{
							scalefreq(0, 1.1f);
							break;
						}
						if(in->state & UI_SUPER_MASK)
						{
							scalefreq(1, 1.1f);
							break;
						}
						if(in->state & UI_ALT_MASK)
						{
							scalefreq(2, 1.1f);
							break;
						}
					}
					else
					{
						if(in->state & UI_CONTROL_MASK)
						{
							shiftphase(0, 0.01);
							break;
						}
						if(in->state & UI_SUPER_MASK)
						{
							shiftphase(1, 0.01);
							break;
						}
						if(in->state & UI_ALT_MASK)
						{
							shiftphase(2, 0.01);
							break;
						}
					}
					/* fall */
				case sc_minus:
				case sc_kp_subtract:
					fmin *= 0.9f;
					fmax *= 0.9f;
					fcenter = (fmin + fmax) * 0.5f;
					scale *= 0.85f;
					correct_frustum();
					break;
				case sc_wheeldown:
					if(in->state & UI_CAPS_LOCK_MASK)
					{
						if(in->state & UI_CONTROL_MASK)
						{
							scalefreq(0, 1/1.1f);
							break;
						}
						if(in->state & UI_SUPER_MASK)
						{
							scalefreq(1, 1/1.1f);
							break;
						}
						if(in->state & UI_ALT_MASK)
						{
							scalefreq(2, 1/1.1f);
							break;
						}
					}
					else
					{
						if(in->state & UI_CONTROL_MASK)
						{
							shiftphase(0, -0.01);
							break;
						}
						if(in->state & UI_SUPER_MASK)
						{
							shiftphase(1, -0.01);
							break;
						}
						if(in->state & UI_ALT_MASK)
						{
							shiftphase(2, -0.01);
							break;
						}
					}
					/* fall */
				case sc_equal:
				case sc_kp_add:
					fmin *= 1.1f;
					fmax *= 1.1f;
					scale *= 1.15f;
					fcenter = (fmin + fmax) * 0.5f;
					correct_frustum();
					break;
				case sc_space:
				case sc_kp_begin:
					if(in->key.press)
					{
						alpha = beta = deltax = deltay = 0;
					}
					break;
				case sc_up:
					deltay -= 0.1f;
					break;
				case sc_kp_up:
					beta += 3;
					break;
				case sc_down:
					deltay += 0.1f;
					break;
				case sc_kp_down:
					beta -= 3;
					break;
				case sc_left:
					deltax += 0.1f;
					break;
				case sc_kp_left:
					alpha -= 3;
					break;
				case sc_right:
					deltax -= 0.1f;
					break;
				case sc_kp_right:
					alpha += 3;
					break;
				case sc_f12:
					agSavePPM("screenshot.ppm");
					break;
			}
			break;
		case ui_wm_event:
			switch(in->configure.wm_event)
			{
				case ui_wm_close:
					return 1;
			}
			break;
		case ui_pointer_motion_event:
			if(in->button & UI_LBUTTON_MASK)
			{
				alpha += in->client_dx;
				beta += in->client_dy;
			}
			if(in->button & UI_RBUTTON_MASK)
			{
				deltax += 0.05f * in->client_dx / scale;
				deltay -= 0.05f * in->client_dy / scale;
			}
			if(in->button & UI_MBUTTON_MASK)
			{
				shiftcenter(in->client_dx, in->client_dy);
			}
			break;
		default:
			break;
	}
	return 0;
}

static void agMainLoop()
{
	ui_event_t ev;
	for(;;)
	{
		display();
		uiWait();
		while(uiPending())
		{
			uiNextEvent(&ev);
			if(handle_event_check_break(&ev))
				return;
		}
	}
}

/*
 * Main
 */

int main(void)
{
    agWindowConfig("MWHEEL+[{CTRL,META,ALT}]+[CAPS], motion + [{L|M|R}]BUTTON", 800, 800, GL_FALSE);
    if(!agInit(AG_REQUIRED_2_0,GL_TRUE,GL_FALSE,GL_FALSE))
    	return -1;
	printf("GLSL version: %s\n",glGetString(GL_SHADING_LANGUAGE_VERSION));

	correct_frustum();
    glDepthFunc(GL_LESS);
    glEnable(GL_DEPTH_TEST);
    NextClearColor();
	
/*
	hack
	for(i = 0;i < 5;i++)
	{
		fmin *= 1.1f;
		fmax *= 1.1f;
		scale *= 1.15f;
	}
*/

	if(installShaders())
    	agMainLoop();
	freeShaders();
   	agShutdown();
    return 0;
}
