/*
 * Cross product of two 3D vectors
 */
void	V_GLOBAL	vec3Cross(vec3_t dst,vec3_t	v1,vec3_t v2)
{
	dst[0]=v1[1]*v2[2]-v1[2]*v2[1];
	dst[1]=v1[2]*v2[0]-v1[0]*v2[2];
	dst[2]=v1[0]*v2[1]-v1[1]*v2[0];
}

/*
 * Returns true if checkpt is inside sphere with center sc and radius sr
 */
vboolean V_GLOBAL
vec3InsideSphere(vec3_t checkpt, vec3_t sc, vec_t sr)
{
	vec3_t d;
	vec_t dist2;
	vec3Sub(d, checkpt, sc);
	dist2 = vec3Dot(d, d);
	return (dist2 < sr * sr) ? true : false;
}

/*
 * Computes plane's equation by given origin and (non-unit) normal.
 */
void	V_GLOBAL
plane3FromOriginNormal(plane3_p dest,vec3_t origin,vec3_t normal)
{
	vec3Unit(dest->normal, normal);
	dest->dist = vec3Dot(dest->normal, origin);
}

/*
 * Computes crossing point of line and plane.
 */
void	V_GLOBAL
vec3FromLinePlane(vec3_t dest, line3_p line, plane3_p plane,vboolean *crossing_ret)
{
	vec_t nd, t;
	nd = vec3Dot(plane->normal, line->dir);
	if(IsZero(nd))
	{
		/* line is parallel to the plane */
		vec_t dist = vec3Dot(plane->normal, line->origin) - plane->dist;
		vec3Copy(dest, line->origin);
		if(crossing_ret)
		{
			if(IsZero(dist))
				*crossing_ret = true;
			else
				*crossing_ret = false;
		}
		return;
	}
	if(crossing_ret)
		*crossing_ret = true;
	t = - (vec3Dot(plane->normal, line->origin) - plane->dist) / nd;
	vec3Scale(dest, line->dir, t);
	vec3Add(dest, dest, line->origin);
}

void	V_GLOBAL
vec3ProjectPointPlane(vec3_t dest, vec3_t pt, plane3_t *plane)
{
	line3_t l;
	vec3Copy(l.origin, pt);
	vec3Copy(l.dir, plane->normal);
	vec3FromLinePlane(dest, &l, plane, NULL);
}

/*
 * Computes equation of line by given two points
 */
void	V_GLOBAL
line3FromPoints(line3_p dest,vec3_t p1,vec3_t p2)
{
	vec3Copy(dest->origin, p1);
	vec3Sub(dest->dir, p2, p1);
	vec3Unit(dest->dir, dest->dir);
}

/*
 * Makes equation of line by direction and origin
 */
void	V_GLOBAL
line3FromDirAndOrigin(line3_p dest,vec3_t dir,vec3_t origin)
{
	vec3Copy(dest->origin, origin);
	vec3Unit(dest->dir, dir);
}

/*
 * Computes equation of plane which is parallel to given lines
 * l1 & l2, and passes throw line l1.
 * Lines must not be parallel or coincident.
 */
void	V_GLOBAL
plane3FromLines(plane3_p dest,line3_p l1,line3_p l2)
{
	vec3Cross(dest->normal, l1->dir, l2->dir);
	vec3Unit(dest->normal, dest->normal);
	dest->dist = vec3Dot(dest->normal, l1->origin);
}

/*
 * Computes common vertical of two lines.
 * In case of coincident lines, both bases contains l1->origin.
 * In case of parallel lines, l1base contains l1->origin.
 */
void
line3CommonVertical(line3_p l1, line3_p l2,
	vec3_t l1base_ret, vec3_t l2base_ret,
	vec_p mindist_ret)
{
	vec_t vn_len;
	plane3_t l1p; /* l1p.normal is the direction of common vertical */
	vec3Cross(l1p.normal, l1->dir, l2->dir);
	vn_len = vec3Abs(l1p.normal);
	if(IsZero(vn_len))
	{
		/*
		 * Lines are parallel or coincident.
		 */
		vec3Copy(l1base_ret, l1->origin);
        /*
         * Compute plane which is perpendicular to both lines:
         *  normal = l1->dir; // or l2->dir
         *  dist = l1base \cdot normal
         */
		vec3Copy(l1p.normal, l1->dir);
		l1p.dist = vec3Dot(l1base_ret, l1p.normal);
		/*
		 * Compute crossing of this plane with l2: project
		 * origin of l2 onto plane.
		 */
		vec3ProjectPointPlane(l2base_ret, l2->origin, &l1p);
		/*
		{
			vec_t t = l1p.dist - vec3Dot(l1p.normal, l2->origin);
			vec3Scale(l2base_ret, l1->dir, t);
			vec3Add(l2base_ret, l2base_ret, l2->origin);
		}
		*/
		if(mindist_ret)
		{
			vec3_t bd;
			vec3Sub(bd, l2base_ret, l1base_ret);
			*mindist_ret = vec3Abs(bd);
		}
	}
	else
	{
		plane3_t l1p2;
		/*
		 * Lines are crossing; compute plane l1p which is parallel to
		 * both l1 and l2 and passes throw line l1.
		 * l1p->normal is 'vn'.
		 */
		vec3Scale(l1p.normal, l1p.normal, 1 / vn_len);
		l1p.dist = vec3Dot(l1p.normal, l1->origin);
		/*
		 * Now we can compute distance between lines (as the distance
		 * between newly constructed plane and any point - l2->origin, for
		 * example - of the second line).
		 * Computing bases of common vertical is more tricky.
		 */
		if(mindist_ret)
			*mindist_ret = (vec_t)fabs(vec3Dot(l1p.normal, l2->origin) - l1p.dist);
		/*
		 * Construct plane l1p2, which is _perpendicular_ to plane l1p
		 * and passes throw line l1.
		 */
		vec3Cross(l1p2.normal, l1->dir, l1p.normal); /* unit */
		l1p2.dist = vec3Dot(l1p2.normal, l1->origin);
		/*
		 * Intersect it with line l2
		 */
		{
			vboolean cross_guard;
			vec3FromLinePlane(l2base_ret, l2, &l1p2, &cross_guard);
			assert(cross_guard);
		}
		/*
		 * Backproject l2base onto l1p
		 */
		vec3ProjectPointPlane(l1base_ret, l2base_ret, &l1p);
		/*
		if(mindist_ret)
		{
			vec_t md;
			vec3_t d;
			vec3Sub(d, l2base_ret, l1base_ret);
			md = *mindist_ret - vec3Abs(d); // should be zero
			assert(IsZero(md));
			printf("%g, %g\n",*mindist_ret,vec3Abs(d));
		}
		*/
	}
}

/*
 * Checks for distance between two line segments
 */
vboolean
segCrossing(vec3_t s1p1, vec3_t s1p2, vec3_t s2p1, vec3_t s2p2,
	vec_p mindist_ret, vec3_t crosspt_ret)
{
#define	HALF	((vec_t)0.5)
	vec3_t s1c, s2c, d, s1base, s2base;
	vec_t s1r, s2r;
	vec_t dist;
	line3_t seg1, seg2;
	/*
	 * Find sircumscribed spheres
	 */
	vec3Sub(s1c, s1p2, s1p1);
	s1r = HALF * vec3Abs(s1c);
	vec3Add(s1c, s1p2, s1p1);
	vec3Scale(s1c, s1c, HALF);
	vec3Sub(s2c, s2p2, s2p1);
	s2r = HALF * vec3Abs(s2c);
	vec3Add(s2c, s2p2, s2p1);
	vec3Scale(s2c, s2c, HALF);
	/*
	 * Check whether or not circumscribed spheres are intersecting
	 */
	vec3Sub(d, s2c, s1c);
	dist = vec3Abs(d);
	/* if distance >= sum of radiuses */
	if(dist >= s1r + s2r)
		/* then can be no intersection anyway (assuming length >> diameter) */
		return false;
	/*
	 * Find bases of common vertical
	 */
	line3FromPoints(&seg1, s1c, s1p2);
	line3FromPoints(&seg2, s2c, s2p2);
	line3CommonVertical(&seg1, &seg2, s1base, s2base, &dist);
	/*
	 * Bases must be inside circumscibed spheres
	 */
	if(!vec3InsideSphere(s1base, s1c, s1r) ||
		!vec3InsideSphere(s2base, s2c, s2r))
		return false;
	if(mindist_ret)
		*mindist_ret = dist;
	vec3Add(crosspt_ret, s1base, s2base);
	vec3Scale(crosspt_ret, crosspt_ret, HALF);
	return true;
#undef HALF
}


/*
 * Draws wireframe box with given corners
 */
void
bboxDraw(vec3_t mins,vec3_t maxs)
{
	glBegin(GL_LINE_LOOP);
		glVertex3f(maxs[0],maxs[1],mins[2]);
		glVertex3f(mins[0],maxs[1],mins[2]);
		glVertex3f(mins[0],mins[1],mins[2]);
		glVertex3f(maxs[0],mins[1],mins[2]);
	glEnd(); /* GL_LINE_LOOP */
	glBegin(GL_LINE_LOOP);
		glVertex3f(maxs[0],maxs[1],maxs[2]);
		glVertex3f(mins[0],maxs[1],maxs[2]);
		glVertex3f(mins[0],mins[1],maxs[2]);
		glVertex3f(maxs[0],mins[1],maxs[2]);
	glEnd(); /* GL_LINE_LOOP */
	glBegin(GL_LINES);
		glVertex3f(maxs[0],maxs[1],mins[2]);
		glVertex3f(maxs[0],maxs[1],maxs[2]);
		glVertex3f(mins[0],maxs[1],mins[2]);
		glVertex3f(mins[0],maxs[1],maxs[2]);
		glVertex3f(mins[0],mins[1],mins[2]);
		glVertex3f(mins[0],mins[1],maxs[2]);
		glVertex3f(maxs[0],mins[1],mins[2]);
		glVertex3f(maxs[0],mins[1],maxs[2]);
	glEnd(); /* GL_LINES */
}

